import logging
from datetime import datetime

import sqlalchemy as sa
from sqlalchemy import text
from sqlalchemy.engine import URL
from sqlalchemy.orm import sessionmaker
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy import event

from tic.mfg.shared import utils as mfg_utils
from tic.shared.utils import db_util as shared_db_utils
from tic.mfg.shared.plugins.db.mfg import (
    models as mfg_models,
    responses as mfg_responses,
    types as mfg_types,
)
from tic.shared.apc.dtos.apc_query import ApcQuery

logger = logging.getLogger()


class MfgRoutines:
    def __init__(
        self,
        host: str = "localhost",
        port: int = 5432,
        database: str = "mfg",
        user: str = "postgres",
        password: str = None,
        is_password_encrypted: bool = True,
        region: str = None,
        schema: str = "dbo",
        ssl_mode: str = None,
        application_name: str = "mfg",
    ):
        self.password = password
        self.schema = None
        self.postgresql_db_autocommit_session = None
        self.postgresql_db_session = None
        self.token_created = datetime.now()
        self.init_routine(
            host,
            port,
            database,
            user,
            password,
            is_password_encrypted,
            region,
            schema,
            ssl_mode,
            application_name,
        )

    def init_routine(
        self,
        host,
        port,
        database,
        user,
        password,
        is_password_encrypted,
        region,
        schema,
        ssl_mode,
        application_name,
    ):
        connect_args = {"options": "-csearch_path={}".format(f"{schema},public")}
        if ssl_mode:
            connect_args["sslmode"] = ssl_mode
            connect_args["application_name"] = application_name
        sql_engine = sa.create_engine(
            URL.create(
                drivername="postgresql",
                host=host,
                port=port,
                database=database,
                password=shared_db_utils.generate_token(host, port, user, region)
                if password is None
                else shared_db_utils.decrypt_password(
                    password, is_password_encrypted, region
                ),
                username=user,
            ),
            connect_args=connect_args,
            pool_size=30,
        )
        mfg_models.Base.metadata.bind = sql_engine
        self.postgresql_db_session = sessionmaker(bind=sql_engine)
        autocommit_sql_engine = sql_engine.execution_options(
            isolation_level="AUTOCOMMIT"
        )
        self.postgresql_db_autocommit_session = sessionmaker(bind=autocommit_sql_engine)

        @event.listens_for(sql_engine, "do_connect")
        def provide_token(dialect, conn_rec, cargs, cparams):
            if self.password is None:
                curr_token = datetime.fromtimestamp(conn_rec.starttime)
                time_elapsed = curr_token - self.token_created
                if time_elapsed.seconds > 900 or "password" not in cparams:
                    self.token_created = datetime.now()
                    cparams["password"] = shared_db_utils.generate_token(
                        host, port, user, region
                    )
                    logger.debug("Token refreshed")

        self.postgresql_db_session().execute(sa.text("SELECT 1"))
        self.schema = schema
        logger.debug("MfgRoutines object initialized successfully!!")

    def get_current_mfg_version(self) -> mfg_responses.MfgVersionResponse:
        session = self.postgresql_db_session()
        query = f"SELECT * from {self.schema}.getcurrentmfgversion()"
        result = session.execute(sa.text(query))
        session.commit()
        if result:
            return mfg_responses.MfgVersionResponse(*result.first())

    def get_mfg_info(self, hashes: list[str]) -> list[mfg_responses.MfgInfoResponse]:
        session = self.postgresql_db_session()
        query = (
            f"SELECT * from {self.schema}.getmfginfo((:sha256)::"
            f"{self.schema}.sha256tabletype[])"
        )
        result = session.execute(
            sa.text(query), mfg_utils.get_sha256_type_array_parameter(hashes)
        )
        session.commit()
        return [mfg_utils.convert_to_mfg_response(row) for row in result]

    def get_malware_tags(
        self, hashes: list[str]
    ) -> list[mfg_responses.Sha256EnumerableStringResponse]:
        session = self.postgresql_db_session()
        query = (
            f"SELECT * from {self.schema}.GetMalwareTags((:sha256)::"
            f"{self.schema}.sha256tabletype[])"
        )
        result = session.execute(
            sa.text(query), mfg_utils.get_sha256_type_array_parameter(hashes)
        )
        session.commit()
        return mfg_utils.convert_to_sha256_list_response(result)

    def get_sources(self, hashes: list[str]) -> list[mfg_responses.SourcesResponse]:
        session = self.postgresql_db_session()
        query = (
            f"SELECT * from {self.schema}.GetSources((:sha256)::"
            f"{self.schema}.sha256tabletype[])"
        )
        result = session.execute(
            sa.text(query), mfg_utils.get_sha256_type_array_parameter(hashes)
        )
        session.commit()
        return mfg_utils.convert_to_source_list_response(result)

    def get_classification_scanners(
        self, hashes: list[str]
    ) -> list[mfg_responses.Sha256EnumerableStringResponse]:
        session = self.postgresql_db_session()
        query = (
            f"SELECT * from "
            f"{self.schema}.GetClassificationScanners((:sha256)::"
            f"{self.schema}.sha256tabletype[])"
        )
        result = session.execute(
            sa.text(query), mfg_utils.get_sha256_type_array_parameter(hashes)
        )
        session.commit()
        return mfg_utils.convert_to_sha256_list_response(result)

    def get_first_and_last_scanned_dates(
        self, hashes: list[str]
    ) -> list[mfg_responses.ScannedDatesResponse]:
        scanned_date_list: list[mfg_responses.ScannedDatesResponse] = list()
        session = self.postgresql_db_session()
        query = (
            f"SELECT * from "
            f"{self.schema}.GetFirstAndLastScannedDate((:sha256)::"
            f"{self.schema}.sha256tabletype[])"
        )
        result = session.execute(
            sa.text(query), mfg_utils.get_sha256_type_array_parameter(hashes)
        )
        session.commit()
        for row in result:
            response = mfg_responses.ScannedDatesResponse()
            response.sha256 = row[0]
            response.first_scanned_date = row[1]
            response.last_scanned_date = row[2]
            scanned_date_list.append(response)
        return scanned_date_list

    def get_total_scans(
        self, hashes: list[str]
    ) -> list[mfg_responses.TotalScansResponse]:
        total_scans: list[mfg_responses.TotalScansResponse] = list()
        session = self.postgresql_db_session()
        query = (
            f"SELECT * from {self.schema}.GetTotalScans((:sha256)::"
            f"{self.schema}.sha256tabletype[])"
        )
        result = session.execute(
            sa.text(query), mfg_utils.get_sha256_type_array_parameter(hashes)
        )
        session.commit()
        for row in result:
            response = mfg_responses.TotalScansResponse()
            response.sha256 = row[0]
            response.total_scans = row[1]
            total_scans.append(response)
        return total_scans

    def get_positive_scanners(
        self, hashes: list[str]
    ) -> list[mfg_responses.Sha256EnumerableStringResponse]:
        session = self.postgresql_db_session()
        query = (
            f"SELECT * from "
            f"{self.schema}.GetPositiveScanners((:sha256)::"
            f"{self.schema}.sha256tabletype[])"
        )
        result = session.execute(
            sa.text(query), mfg_utils.get_sha256_type_array_parameter(hashes)
        )
        session.commit()
        return mfg_utils.convert_to_sha256_list_response(result)

    def get_changes_to_be_updated(self, hashes: list[str, int]) -> list[str, int]:
        session = self.postgresql_db_session()
        hash_list = [mfg_utils.hash_to_binary(_hash[0]) for _hash in hashes]
        result = (
            session.query(mfg_models.UpdateHashes)
            .filter(mfg_models.UpdateHashes.sha256.in_(hash_list))
            .all()
        )
        session.commit()
        hashes_in_db = {
            mfg_utils.binary_to_hash(row.sha256): row.update_hash for row in result
        }
        return [
            _hash
            for _hash in hashes
            if (
                _hash[0].lower() not in hashes_in_db
                or hashes_in_db[_hash[0].lower()] != _hash[1]
            )
        ]

    def get_processing_checkpoints(self) -> mfg_responses.ProcessingCheckpoints:
        session = self.postgresql_db_session()
        processing_checkpoints = mfg_responses.ProcessingCheckpoints()
        processing_checkpoints.last_processed_id = (
            session.query(mfg_models.FeedLog)
            .order_by(mfg_models.FeedLog.id.desc())
            .first()
            .mfg_last_processed_id
        )
        processing_checkpoints.newest_update_id = (
            session.query(mfg_models.HashInsertsAndUpdate)
            .order_by(mfg_models.HashInsertsAndUpdate.id.desc())
            .first()
            .id
        )
        session.commit()
        return processing_checkpoints

    def get_mfg_updates(self, lower_bound, upper_bound) -> list[str]:
        session = self.postgresql_db_session()
        result = (
            session.query(mfg_models.HashInsertsAndUpdate)
            .with_entities(mfg_models.HashInsertsAndUpdate.sha256)
            .join(
                mfg_models.HashIndex,
                mfg_models.HashInsertsAndUpdate.sha256 == mfg_models.HashIndex.sha256,
            )
            .join(mfg_models.Mfg, mfg_models.HashIndex.mfg_id == mfg_models.Mfg.id)
            .filter(mfg_models.HashInsertsAndUpdate.id <= upper_bound)
            .filter(mfg_models.HashInsertsAndUpdate.id > lower_bound)
            .all()
        )
        session.commit()
        return [mfg_utils.binary_to_hash(row.sha256.lower()) for row in set(result)]

    def update_feed_log(self, mfg_last_processed_id: int, bit9_last_processed_id: int):
        session = self.postgresql_db_session()
        feed_log = mfg_models.FeedLog(
            mfg_last_processed_id=mfg_last_processed_id,
            bit9_last_processed_id=-bit9_last_processed_id,
        )
        session.add(feed_log)
        session.commit()

    def update_mfg_scanners(self, mfg_scanners: list[mfg_types.MfgScannerTableType]):
        session = self.postgresql_db_session()
        query = (
            f"CALL {self.schema}.UpdateMfgScanners((:mfg_scanners)::"
            f"{self.schema}.mfgscannertabletype[])"
        )
        session.execute(
            sa.text(query),
            dict(mfg_scanners=[scanner.tuple() for scanner in mfg_scanners]),
        )
        session.commit()

    def update_unofficial_malware_tags(
        self, malware_tags: list[mfg_types.UpdateMalwareTagsTableType]
    ):
        session = self.postgresql_db_session()
        query = (
            f"CALL {self.schema}.UpdateUnofficialMalwareTags("
            f"(:malware_tags)::{self.schema}.updatemalwaretagstabletype[])"
        )
        session.execute(
            sa.text(query), dict(malware_tags=[tag.tuple() for tag in malware_tags])
        )
        session.commit()

    def insert_update_hashes(self, all_updated_data: list[(str, int)]):
        session = self.postgresql_db_session()
        query = (
            f"SELECT * from {self.schema}.InsertUpdateHashes((:sha256)::"
            f"{self.schema}.updatehashestabletype[])"
        )
        session.execute(
            sa.text(query),
            dict(
                sha256=[
                    (mfg_utils.hash_to_binary(sha256[0]), sha256[1])
                    for sha256 in all_updated_data
                ]
            ),
        )
        session.commit()

    def insert_update_mfg(self, mfg_updates: list[mfg_types.MfgTableType]):
        session = self.postgresql_db_session()
        query = (
            f"CALL {self.schema}.InsertUpdateMfg((:mfg_updates)::"
            f"{self.schema}.mfgtabletype[])"
        )
        session.execute(
            sa.text(query), dict(mfg_updates=[update.tuple() for update in mfg_updates])
        )
        session.commit()

    def insert_apc_hashes(self, hashes: list[ApcQuery]):
        hashes_to_insert = [
            {
                "sha256": value.sha256,
                "verdict": value.category,
                "analysis_date": value.analysis_date,
            }
            for value in hashes
        ]
        session = self.postgresql_db_session()
        query = insert(mfg_models.ApcHashes).values(hashes_to_insert)
        query = query.on_conflict_do_update(
            index_elements=["sha256"],
            where=(
                (mfg_models.ApcHashes.analysis_date.is_(None))
                | (mfg_models.ApcHashes.analysis_date < query.excluded.analysis_date)
            ),
            set_={
                "last_updated_ts": datetime.now(),
                "verdict": query.excluded.verdict,
                "analysis_date": query.excluded.analysis_date,
            },
        )
        session.execute(query)
        session.commit()

    def delete_apc_hashes(self, retention_months):
        query = text(
            """
            delete from apc_hashes ah
            where ah.last_updated_ts < (now() - ':retention_months months'::interval)
            and exists(
            select 1 from hash_index hi where hi.sha256 = decode(ah.sha256, 'hex')
            )
        """
        )
        session = self.postgresql_db_session()
        session.execute(query, {"retention_months": retention_months})
        session.commit()
        self.reindex_table("apc_hashes")

    def reindex_table(self, table_name):
        session = self.postgresql_db_autocommit_session()
        session.execute(text(f"REINDEX TABLE CONCURRENTLY {table_name}"))
